from pydantic import EmailStr, Field
from functools import lru_cache
import importlib.metadata
import unipoll_api.__version__ as version_file
from pydantic_settings import BaseSettings, SettingsConfigDict


try:
    VERSION = importlib.metadata.version('unipoll-api')
except importlib.metadata.PackageNotFoundError:
    VERSION = version_file.version


class Settings(BaseSettings):  # type: ignore
    app_name: str = Field(default="University Polling API",
                          title="App Name", description="The name of the API.")
    app_version: str = Field(default=VERSION, title="App Version", description="The version of the API.")
    app_description: str = Field(default=("An Open Source API for creating surveys and polls "
                                          "to assist in university research."),
                                 title="App Description", description="A description of the API.")
    admin_email: EmailStr = Field(default="admin@example.com",
                                  title="Admin Email", description="The email address of the API administrator.")
    mongodb_url: str = Field(default="mongodb://localhost:27017",
                             title="MongoDB URL", description="The URL of the MongoDB database.")
    secrete_key: str = Field(default="secret", title="Secrete Key", description="The secrete key of the API.")
    origins: str = "*"
    host: str = "0.0.0.0"
    port: int = 9000
    reload: bool = True
    model_config = SettingsConfigDict(env_file=".env")


@lru_cache()
def get_settings() -> Settings:
    return Settings()



from unipoll_api.exceptions import resource
from unipoll_api.documents import Poll, ResourceID, Account


# Exception for when a Poll with the same name already exists
class NonUniqueName(resource.NonUniqueName):
    def __init__(self, poll: Poll):
        super().__init__("Poll", resource_name=poll.name)


# Exception for when an error occurs during Poll creation
class ErrorWhileCreating(resource.ErrorWhileCreating):
    def __init__(self, poll: Poll):
        super().__init__("Poll", resource_name=poll.name)


# Exception for when a Poll is not found
class PollNotFound(resource.ResourceNotFound):
    def __init__(self, poll_id: ResourceID):
        super().__init__("Poll", resource_id=poll_id)


# Not authorized
class UserNotAuthorized(resource.UserNotAuthorized):
    def __init__(self, account: Account, poll: Poll, action: str):
        super().__init__(account, f'poll {poll.name}', action)


# Action not found
class ActionNotFound(resource.ActionNotFound):
    def __init__(self, action: str):
        super().__init__('poll', action)



from unipoll_api.exceptions import resource
from unipoll_api.documents import ResourceID


class AccountNotFound(resource.ResourceNotFound):
    def __init__(self, account_id: ResourceID):
        super().__init__("Account", resource_id=account_id)


class ErrorWhileDeleting(resource.ErrorWhileDeleting):
    def __init__(self, account_id: ResourceID):
        super().__init__("Account", resource_id=account_id)



from datetime import datetime, timezone
from typing import (
    Any,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
)

from beanie import Document, PydanticObjectId
from fastapi_users.authentication.strategy.db import AccessTokenDatabase
from pydantic import BaseModel, Field
from pymongo import IndexModel
from beanie.odm.enums import SortDirection


class BeanieBaseAccessToken(BaseModel):
    access_token: str
    refresh_token: str
    user_id: PydanticObjectId
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))

    class Settings:
        indexes = [IndexModel("access_token", unique=True),
                   IndexModel("refresh_token", unique=True)]


class BeanieBaseAccessTokenDocument(BeanieBaseAccessToken, Document):  # type: ignore
    pass


AP_BEANIE = TypeVar("AP_BEANIE", bound=BeanieBaseAccessTokenDocument)


class BeanieAccessTokenDatabase(Generic[AP_BEANIE], AccessTokenDatabase[AP_BEANIE]):  # type: ignore
    """
    Access token database adapter for Beanie.

    :param access_token_model: Beanie access token model.
    """

    def __init__(self, access_token_model: Type[AP_BEANIE]):
        self.access_token_model = access_token_model

    async def get_by_token(self, token: str,
                           max_age: Optional[datetime] = None) -> Optional[AP_BEANIE]:
        query: Dict[str, Any] = {"access_token": token}
        if max_age is not None:
            query["created_at"] = {"$gte": max_age}
        res = await self.access_token_model.find_one(query)
        return res

    async def get_by_refresh_token(self, token: str) -> Optional[AP_BEANIE]:
        query: Dict[str, Any] = {"refresh_token": token}
        res = await self.access_token_model.find_one(query)
        return res

    # Returns find query (not a document)
    async def get_token_family_by_user_id(self, user_id: PydanticObjectId):
        access_token = self.access_token_model.find({"user_id": user_id},
                                                    sort=[("created_at", SortDirection.DESCENDING)])
        return access_token

    async def create(self, create_dict: Dict[str, Any]) -> AP_BEANIE:
        access_token = self.access_token_model(**create_dict)
        await access_token.create()
        return access_token

    async def update(
        self, access_token: AP_BEANIE, update_dict: Dict[str, Any]
    ) -> AP_BEANIE:
        for key, value in update_dict.items():
            setattr(access_token, key, value)
        await access_token.save()  # type: ignore
        return access_token

    async def delete(self, access_token: AP_BEANIE) -> None:
        await access_token.delete()  # type: ignore



from unipoll_api import AccountManager
from unipoll_api.documents import Account, Workspace, Group, Policy, Resource, Member
from unipoll_api.schemas import MemberSchemas, PolicySchemas, GroupSchemas
from unipoll_api.exceptions import ResourceExceptions
from unipoll_api.utils import Permissions
from unipoll_api.utils.permissions import check_permissions
from unipoll_api.dependencies import get_member


# Helper function to get policies from a resource
# NOTE: This can be moved to utils
async def get_policies_from_resource(resource: Resource) -> list[Policy]:
    policies: list[Policy] = []
    try:
        await check_permissions(resource, "get_policies")
        return resource.policies  # type: ignore
    except ResourceExceptions.UserNotAuthorized:
        print("User not authorized")
        account = AccountManager.active_user.get()
        member = await get_member(account, resource)
        for policy in resource.policies:
            if policy.policy_holder.ref.id == member.id:  # type: ignore
                policies.append(policy)  # type: ignore
        return policies


# Get all policies of a workspace
async def get_policies(policy_holder: Member | Group | None = None,
                       resource: Resource | None = None) -> PolicySchemas.PolicyList:
    policy_list = []
    policy: Policy
    all_policies = []

    # Get policies from a specific resource
    if resource:
        all_policies = await get_policies_from_resource(resource)
    # Get policies from all resources
    else:
        all_workspaces = Workspace.find(fetch_links=True)
        all_groups = Group.find(fetch_links=True)
        all_resources = await all_workspaces.to_list() + await all_groups.to_list()

        for resource in all_resources:
            all_policies += await get_policies_from_resource(resource)

    # Build policy list
    for policy in all_policies:
        # Filter by policy_holder if specified
        if policy_holder:
            if (policy.policy_holder.ref.id != policy_holder.id):
                continue
        policy_list.append(await get_policy(policy, False))
    # Return policy list
    return PolicySchemas.PolicyList(policies=policy_list)


async def get_policy(policy: Policy, permission_check: bool = True) -> PolicySchemas.PolicyShort:
    # Get the parent resource of the policy
    parent_resource = await policy.get_parent_resource(fetch_links=True)
    await check_permissions(parent_resource, "get_policies", permission_check)

    # Get the policy holder
    policy_holder = await policy.get_policy_holder()
    member, group = None, None
    if policy_holder.get_document_type() == "Member":
        await policy_holder.fetch_link("account")
        account: Account = policy_holder.account  # type: ignore
        member = MemberSchemas.Member(id=policy_holder.id,
                                      account_id=account.id,
                                      email=account.email,
                                      first_name=account.first_name,
                                      last_name=account.last_name)
    elif policy_holder.get_document_type() == "Group":
        group = GroupSchemas.Group(id=policy_holder.id,
                                   name=policy_holder.name,
                                   description=policy_holder.description)

    # Get the permissions based on the resource type and convert it to a list of strings
    permission_type = Permissions.PermissionTypes[parent_resource.get_document_type()]
    permissions = permission_type(policy.permissions).name.split('|')  # type: ignore

    # Return the policy
    return PolicySchemas.PolicyShort(id=policy.id,
                                     policy_holder_type=policy.policy_holder_type,
                                     policy_holder=member or group,
                                     permissions=permissions)


async def update_policy(policy: Policy,
                        new_permissions: list[str],
                        check_permissions: bool = True) -> PolicySchemas.PolicyOutput:

    parent_resource = await policy.get_parent_resource(fetch_links=True)

    # Check if the user has the required permissions to update the policy
    await Permissions.check_permissions(parent_resource, "update_policies", check_permissions)
    permission_type = Permissions.PermissionTypes[parent_resource.get_document_type()]

    # Calculate the new permission value from request
    new_permission_value = 0
    for i in new_permissions:
        try:
            new_permission_value += permission_type[i].value
        except KeyError:
            raise ResourceExceptions.InvalidPermission(i)
    # Update permissions
    policy.permissions = permission_type(new_permission_value)
    await Policy.save(policy)

    policy_holder = await policy.get_policy_holder()
    member, group = None, None
    if policy_holder.get_document_type() == "Member":
        await policy_holder.fetch_link("account")
        account: Account = policy_holder.account  # type: ignore
        member = MemberSchemas.Member(id=policy_holder.id,
                                      account_id=account.id,
                                      email=account.email,
                                      first_name=account.first_name,
                                      last_name=account.last_name)
    elif policy_holder.get_document_type() == "Group":
        group = GroupSchemas.Group(id=policy_holder.id,
                                   name=policy_holder.name,
                                   description=policy_holder.description)

    return PolicySchemas.PolicyOutput(permissions=permission_type(policy.permissions).name.split('|'),  # type: ignore
                                      policy_holder=member or group)



from fastapi import WebSocket


class WebSocketManager:
    def __init__(self) -> None:
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket) -> None:
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket) -> None:
        self.active_connections.remove(websocket)

    async def send_personal_message(self, message: str, websocket: WebSocket) -> None:
        await websocket.send_text(message)

    async def broadcast(self, message: str) -> None:
        for connection in self.active_connections:
            await connection.send_text(message)



from beanie import WriteRules
from beanie.operators import In
from unipoll_api.documents import Account, Group, ResourceID, Workspace, Member
from unipoll_api.utils import Permissions
from unipoll_api.schemas import MemberSchemas
# from unipoll_api import AccountManager
from unipoll_api.exceptions import ResourceExceptions
from unipoll_api.dependencies import get_member


async def get_members(resource: Workspace | Group, check_permissions: bool = True) -> MemberSchemas.MemberList:
    # Check if the user has permission to add members
    await Permissions.check_permissions(resource, "get_members", check_permissions)

    def build_member_scheme(member: Member) -> MemberSchemas.Member:
        account: Account = member.account  # type: ignore
        return MemberSchemas.Member(id=member.id,
                                    account_id=account.id,
                                    first_name=account.first_name,
                                    last_name=account.last_name,
                                    email=account.email)

    member_list = [build_member_scheme(member) for member in resource.members]  # type: ignore
    # Return the list of members
    return MemberSchemas.MemberList(members=member_list)


# Add groups/members to group
async def add_members(resource: Workspace | Group,
                      account_id_list: list[ResourceID],
                      check_permissions: bool = True) -> MemberSchemas.MemberList:
    # Check if the user has permission to add members
    await Permissions.check_permissions(resource, "add_members", check_permissions)

    # Remove duplicates from the list of accounts
    accounts = set(account_id_list)
    # Remove existing members from the accounts set
    accounts = accounts.difference({member.id for member in resource.members})  # type: ignore
    # Find the accounts from the database
    account_list = await Account.find(In(Account.id, accounts)).to_list()
    # Add the accounts to the group member list with basic permissions

    new_members = []

    for account in account_list:
        default_permissions = eval("Permissions." + resource.get_document_type().upper() + "_BASIC_PERMISSIONS")
        if resource.get_document_type() == "Group":
            member = await get_member(account, resource.workspace)  # type: ignore
            new_member = await resource.add_member(member, default_permissions, save=False)
            new_members.append(new_member)
        elif resource.get_document_type() == "Workspace":
            new_member = await resource.add_member(account, default_permissions, save=False)
            new_members.append(new_member)
    await resource.save(link_rule=WriteRules.WRITE)  # type: ignore

    member_list = []
    for new_member in new_members:
        account: Account = new_member.account  # type: ignore
        member_list.append(MemberSchemas.Member(id=new_member.id,
                                                account_id=account.id,
                                                first_name=account.first_name,
                                                last_name=account.last_name,
                                                email=account.email))

    # Return the list of members added to the group
    return MemberSchemas.MemberList(members=member_list)


# Remove a member from a workspace
async def remove_member(resource: Workspace | Group,
                        member: Member,
                        permission_check: bool = True) -> MemberSchemas.MemberList:
    # Check if the user has permission to add members
    await Permissions.check_permissions(resource, "remove_members", permission_check)

    # Check if the account is a member of the workspace
    if member.id not in [ResourceID(member.id) for member in resource.members]:  # type: ignore
        raise ResourceExceptions.ResourceNotFound("Member", member.id)

    # Remove the account from the workspace/group
    if await resource.remove_member(member):
        # Return the list of members added to the group
        member_list = [MemberSchemas.Member(**account.model_dump()) for account in resource.members]  # type: ignore
        return MemberSchemas.MemberList(members=member_list)
    raise ResourceExceptions.ErrorWhileRemovingMember(resource, member)



from unipoll_api.documents import Account, Workspace, Group, AccessToken
from unipoll_api.account_manager import active_user
from unipoll_api.exceptions import AccountExceptions


# Delete account
async def delete_account(account: Account | None = None) -> None:
    if not account:
        account = active_user.get()

    # Delete account
    await Account.delete(account)

    # Delete all policies associated with account
    # BUG: This doesn't work due to type mismatch
    # await Policy.find({"policy_holder": account}).delete()  # type: ignore

    # Remove account from all workspaces
    workspaces = await Workspace.find(Workspace.members.id == account.id).to_list()  # type: ignore
    for workspace in workspaces:
        await workspace.remove_member(account)  # type: ignore

    # Remove account from all groups
    groups = await Group.find(Group.members.id == account.id).to_list()  # type: ignore
    for group in groups:
        await group.remove_member(account)  # type: ignore

    # Check if account was deleted
    if await Account.get(account.id):  # type: ignore
        raise AccountExceptions.ErrorWhileDeleting(account.id)  # type: ignore

    # Delete access tokens associated with account
    await AccessToken.find(AccessToken.user_id == account.id).delete()  # type: ignore



from typing import Optional
from pydantic import BaseModel


class LoginResponse(BaseModel):
    access_token: str
    token_type: str = "Bearer"
    scope: Optional[str] = None
    client_id: Optional[str] = None
    expires_in: int = 3600
    refresh_token: str


class PostmanRefreshTokenRequest(BaseModel):
    refresh_token: str
    grant_type: str = "refresh_token"



# from typing import ForwardRef, NewType, TypeAlias, Optional
from typing import Literal
from bson import DBRef
from beanie import Document as BeanieDocument
from beanie import BackLink, WriteRules, after_event, Insert, Link, PydanticObjectId  # BackLink
from fastapi_users_db_beanie import BeanieBaseUser
from pydantic import Field
from unipoll_api.utils import colored_dbg as Debug
from unipoll_api.utils.token_db import BeanieBaseAccessToken


# Document
class Document(BeanieDocument):
    @classmethod
    def get_document_type(cls) -> str:
        return cls._document_settings.name  # type: ignore


# Create a link to the Document model
async def create_link(document: Document) -> Link:
    ref = DBRef(collection=document._document_settings.name,  # type: ignore
                id=document.id)
    link = Link(ref, type(document))
    return link


# Custom PydanticObjectId class to override due to a bug
class ResourceID(PydanticObjectId):
    @classmethod
    def __modify_schema__(cls, field_schema):  # type: ignore
        field_schema.update(
            type="string",
            example="5eb7cf5a86d9755df3a6c593",
        )


class AccessToken(BeanieBaseAccessToken, Document):  # type: ignore
    pass


class Resource(Document):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    name: str = Field(
        title="Name", description="Name of the resource", min_length=3, max_length=50)
    description: str = Field(default="", title="Description", max_length=1000)
    policies: list[Link["Policy"]] = []

    @after_event(Insert)
    def create_group(self) -> None:
        Debug.info(f'New {self.get_document_type()} "{self.id}" has been created')

    async def add_policy(self, policy_holder: "Group | Member", permissions, save: bool = True) -> "Policy":
        new_policy = Policy(policy_holder_type=policy_holder.get_document_type(),  # type: ignore
                            policy_holder=(await create_link(policy_holder)),
                            permissions=permissions,
                            parent_resource=(await create_link(self)))  # type: ignore

        # Add the policy to the group
        self.policies.append(new_policy)  # type: ignore
        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return new_policy

    async def remove_policy(self, policy: "Policy", save: bool = True) -> None:
        for i, p in enumerate(self.policies):
            if policy.id == p.ref.id:
                self.policies.remove(p)
                if save:
                    await self.save(link_rule=WriteRules.WRITE)  # type: ignore

    async def remove_policy_by_holder(self, policy_holder: "Group | Member", save: bool = True) -> None:
        for policy in self.policies:
            if policy.policy_holder.ref.id == policy_holder.id:  # type: ignore
                self.policies.remove(policy)
                if save:
                    await self.save(link_rule=WriteRules.WRITE)  # type: ignore


class Account(BeanieBaseUser, Document):  # type: ignore
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    first_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    last_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")


class Workspace(Resource):
    members: list[Link["Member"]] = []
    groups: list[Link["Group"]] = []
    polls: list[Link["Poll"]] = []

    async def add_member(self, account: "Account", permissions, save: bool = True) -> "Member":
        new_member = await Member(account=account, resource=(await create_link(self))).create()  # type: ignore
        new_policy = await self.add_policy(new_member, permissions, save=False)  # type: ignore
        new_member.policies.append(new_policy)  # type: ignore

        self.members.append(new_member)  # type: ignore

        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return new_member

    async def remove_member(self, member_to_delete: "Member", save: bool = True) -> bool:
        # Remove the account from the workspace
        for member in self.members:
            if member.id == member_to_delete.id:  # type: ignore
                self.members.remove(member)
                await member.delete()  # type: ignore
                # type: ignore
                Debug.info(f"Removed member {member.id} from {self.get_document_type()} {self.id}")  # type: ignore
                break

        # Remove the policy from the workspace
        await self.remove_policy_by_holder(member_to_delete, save=False)  # type: ignore

        # Remove the member from all groups in the workspace
        group: Group
        for group in self.groups:  # type: ignore
            await group.remove_member(member_to_delete, save=False)
            await group.remove_policy_by_holder(member_to_delete, save=False)
            await Group.save(group, link_rule=WriteRules.WRITE)

        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return True


class Group(Resource):
    workspace: BackLink[Workspace] = Field(original_field="groups")
    members: list[Link["Member"]] = []
    groups: list[Link["Group"]] = []

    async def add_member(self, member: "Member", permissions, save: bool = True) -> "Member":
        if member.workspace.id != self.workspace.id:  # type: ignore
            from unipoll_api.exceptions import WorkspaceExceptions
            raise WorkspaceExceptions.UserNotMember(
                self.workspace, member)  # type: ignore

        # Add the member to the group's list of members
        self.members.append(member)  # type: ignore
        # Create a policy for the new member
        await self.add_policy(member, permissions, save=False)  # type: ignore
        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return member

    async def remove_member(self, member: "Member", save: bool = True) -> bool:
        # Remove the account from the group
        for _member in self.members:
            if _member.id == member.id:  # type: ignore
                self.members.remove(_member)
                # type: ignore
                Debug.info(
                    f"Removed member {member.id} from {self.get_document_type()} {self.id}")  # type: ignore
                break

        # Remove the policy from the group
        await self.remove_policy_by_holder(member, save=False)  # type: ignore

        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return True


class Poll(Resource):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    workspace: Link[Workspace]
    public: bool
    published: bool
    questions: list
    policies: list[Link["Policy"]]


class Policy(Document):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    parent_resource: Link[Workspace] | Link[Group] | Link[Poll]
    policy_holder_type: Literal["Member", "Group"]
    policy_holder: Link["Group"] | Link["Member"]
    permissions: int

    async def get_parent_resource(self, fetch_links: bool = False) -> Workspace | Group | Poll:
        from unipoll_api.exceptions.resource import ResourceNotFound
        collection = eval(self.parent_resource.ref.collection)
        parent: Workspace | Group | Poll = await collection.get(self.parent_resource.ref.id,
                                                                fetch_links=fetch_links)
        if not parent:
            ResourceNotFound(self.parent_resource.ref.collection,
                             self.parent_resource.ref.id)
        return parent

    async def get_policy_holder(self, fetch_links: bool = False) -> "Group | Member":
        from unipoll_api.exceptions.policy import PolicyHolderNotFound
        collection = eval(self.policy_holder.ref.collection)
        policy_holder: Group | Member = await collection.get(self.policy_holder.ref.id,
                                                             fetch_links=fetch_links)
        if not policy_holder:
            PolicyHolderNotFound(self.policy_holder.ref.id)
        return policy_holder


class Member(Document):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    account: Link[Account]
    workspace: BackLink[Workspace] = Field(original_field="members")
    groups: list[BackLink[Group]] = Field(original_field="members")
    policies: list[Link[Policy]] = []



import json
import uvicorn
import os
import argparse
from fastapi import FastAPI
from fastapi.routing import APIRoute
from fastapi.middleware.cors import CORSMiddleware
from beanie import init_beanie
from unipoll_api.routes import router
from unipoll_api.mongo_db import mainDB, documentModels
from unipoll_api.config import get_settings
from unipoll_api.__version__ import version
from unipoll_api.utils import cli_args, colored_dbg


# Apply setting from configuration file
settings = get_settings()

# Create FastAPI application
app = FastAPI(
    title=settings.app_name,               # Title of the application
    description=settings.app_description,  # Description of the application
    version=settings.app_version,          # Version of the application
)

# Add endpoints defined in the routes directory
app.include_router(router)

# Add CORS middleware to allow cross-origin requests
origins = settings.origins.split(",")

# Middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"]
)


# Initialize Mongo Database on startup
@app.on_event("startup")
async def on_startup() -> None:
    # Simplify operation IDs so that generated API clients have simpler function names
    # Each route will have its operation ID set to the method name
    for route in app.routes:
        if isinstance(route, APIRoute):
            route.operation_id = route.name

    await init_beanie(
        database=mainDB,  # type: ignore
        document_models=documentModels  # type: ignore
    )


# Run the application
def start_server(host: str = settings.host, port: int = settings.port, reload: bool = settings.reload):
    uvicorn.run('unipoll_api.app:app', reload=reload, host=host, port=port)


# Check if IP address is valid
def check_ip(arg_value):
    address = arg_value.split(".")
    if len(address) != 4:
        raise argparse.ArgumentTypeError("invalid host value")
    for i in address:
        if int(i) > 255 or int(i) < 0:
            raise argparse.ArgumentTypeError("invalid host value")
    return arg_value


def cli_entry_point():
    args = cli_args.parse_args()

    if args.command == "run":
        run(args.host, args.port, args.reload)
    elif args.command == "setup":
        setup()
    elif args.command == "get-openapi":
        get_openapi()
    else:
        print("Invalid command")


def run(host=settings.host, port=settings.port, reload=settings.reload):
    # args = run_parser.parse_args()
    colored_dbg.info("University Polling API v{}".format(version))
    start_server(host, port, reload)


def setup():
    # Print current directory
    # print("Current directory: {}".format(os.getcwd()))

    # Get user input
    host = input("Host IP address [{}]: ".format(settings.host))
    port = input("Host port number [{}]: ".format(settings.port))
    mongodb_url = input("MongoDB URL [{}]: ".format(settings.mongodb_url))
    origins = input("Origins [{}]: ".format(settings.origins))
    admin_email = input("Admin email [{}]: ".format(settings.admin_email))

    # Write to .env file
    with open(".env", "w") as f:
        f.write("HOST={}\n".format(host if host else settings.host))
        f.write("PORT={}\n".format(port if port else settings.port))
        f.write("MONGODB_URL={}\n".format(mongodb_url if mongodb_url else settings.mongodb_url))
        f.write("ORIGINS={}\n".format(origins if origins else settings.origins))
        f.write("ADMIN_EMAIL={}\n".format(admin_email if admin_email else settings.admin_email))

    # Print success message
    print(f"Your configuration has been saved to {os.getcwd()}/.env")


def get_openapi():
    if not app.openapi_schema:
        openapi_schema = app.openapi()
        app.openapi_schema = openapi_schema
    json.dump(app.openapi_schema, open("openapi.json", "w"), indent=2)

    # Print success message
    print(f"OpenAPI schema saved to {os.getcwd()}/openapi.json")



# Handle WebSocket connections
from fastapi import APIRouter, Depends, WebSocket, WebSocketDisconnect
from unipoll_api.websocket_manager import WebSocketManager
from unipoll_api.dependencies import websocket_auth


router: APIRouter = APIRouter()

# Create a connection manager to manage WebSocket connections
manager = WebSocketManager()


@router.websocket("")
async def open_websocket_endpoint(websocket: WebSocket, auth: dict = Depends(websocket_auth)):
    await manager.connect(websocket)
    print("auth: ", auth)
    if auth["cookie"]:
        print("cookie: ", auth["cookie"])
        # account_id = AccessToken.find(user_id=auth.token)
    elif auth["token"]:
        print("token: ", auth["token"])
    else:
        print("no auth")
    try:
        while True:
            data = await websocket.receive_text()
            await manager.send_personal_message(f"You wrote: {data}", websocket)
    except WebSocketDisconnect:
        manager.disconnect(websocket)



from typing import Optional, Any
from pydantic import ConfigDict, BaseModel
from unipoll_api.documents import ResourceID
from unipoll_api.schemas.question import Question


class PollResponse(BaseModel):
    id: Optional[ResourceID] = None
    # workspace: Optional[Union['Workspace', 'WorkspaceShort']]
    workspace: Optional[Any] = None
    name: str
    description: str
    public: bool
    published: bool
    questions: Optional[list[Question]] = None
    policies: Optional[list] = None
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "id": "1a2b3c4d5e6f7g8h9i0j",
            "name": "Poll 01",
            "description": "This is an example poll",
            "published": True
        }
    })


class PollShort(BaseModel):
    id: ResourceID
    name: str
    description: str
    public: bool
    published: bool
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "poll": {
                "id": "1a2b3c4d5e6f7g8h9i0j",
                "name": "Poll 01",
                "description": "This is an example poll",
                "published": True
            }
        }
    })


class PollList(BaseModel):
    polls: list[PollShort]
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "polls": [
                {
                    "id": "1a2b3c4d5e6f7g8h9i0j",
                    "name": "Poll 01",
                    "description": "This is an example poll",
                    "published": True
                },
                {
                    "id": "1a2b3c4d5e6f7g8h9i0j",
                    "name": "Poll 02",
                    "description": "This is an example poll",
                    "published": True
                }
            ]
        }
    })


class CreatePollRequest(BaseModel):
    name: str
    description: str
    public: bool
    published: bool
    questions: list[Question]


class UpdatePollRequest(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    public: Optional[bool] = None
    published: Optional[bool] = None
    questions: Optional[list[Question]] = None
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "Poll 01",
            "description": "This is an example poll",
            "published": True
        }
    })


# Forward references
from unipoll_api.schemas.workspace import Workspace, WorkspaceShort   # noqa: E402
Workspace.model_rebuild()
WorkspaceShort.model_rebuild()



from beanie import WriteRules
from beanie.operators import Or
from bson import DBRef

from unipoll_api import AccountManager
from unipoll_api.documents import Policy, Workspace, Group, Account
from unipoll_api import actions
from unipoll_api.schemas import GroupSchemas, WorkspaceSchemas
from unipoll_api.exceptions import GroupExceptions, WorkspaceExceptions, ResourceExceptions
from unipoll_api.utils import Permissions
from unipoll_api.dependencies import get_member


# Get list of groups
async def get_groups(workspace: Workspace | None = None,
                     account: Account | None = None,
                     name: str | None = None) -> GroupSchemas.GroupList:
    account = account or AccountManager.active_user.get()

    search_filter = {}
    if name:
        search_filter['name'] = name  # type: ignore
    if workspace:
        search_filter['workspace._id'] = workspace.id  # type: ignore
    if account:
        search_filter['members.account._id'] = account.id  # type: ignore
    search_result = await Group.find(search_filter, fetch_links=True).to_list()

    # TODO: Rewrite to iterate over list of workspaces
    # TODO: to avoid permission check for every group if the user has permission to get all groups

    groups = []
    for group in search_result:
        try:
            groups.append((await get_group(group=group)).model_dump(exclude_none=True))
        except Exception:
            pass

    return GroupSchemas.GroupList(groups=groups)


# Create a new group with account as the owner
async def create_group(workspace: Workspace,
                       name: str,
                       description: str,
                       check_permissions: bool = True) -> GroupSchemas.GroupCreateOutput:

    await Permissions.check_permissions(workspace, "add_groups", check_permissions)
    account = AccountManager.active_user.get()

    member = await get_member(account, workspace)

    # Check if group name is unique
    group: Group  # For type hinting, until Link type is supported
    for group in workspace.groups:  # type: ignore
        if group.name == name:
            raise GroupExceptions.NonUniqueName(group)

    # Create a new group
    new_group = Group(name=name,
                      description=description,
                      workspace=workspace)  # type: ignore

    # Check if group was created
    if not new_group:
        raise GroupExceptions.ErrorWhileCreating(new_group)

    # Add the account to group member list
    await new_group.add_member(member, Permissions.GROUP_ALL_PERMISSIONS)

    # Create a policy for the new group
    await workspace.add_policy(new_group, Permissions.WORKSPACE_BASIC_PERMISSIONS, False)
    workspace.groups.append(new_group)  # type: ignore
    await Workspace.save(workspace, link_rule=WriteRules.WRITE)

    # Return the new group
    return GroupSchemas.GroupCreateOutput(**new_group.model_dump(include={'id', 'name', 'description'}))


# Get group
async def get_group(group: Group,
                    include_members: bool = False,
                    include_policies: bool = False,
                    check_permissions: bool = True) -> GroupSchemas.Group:
    try:
        await Permissions.check_permissions(group.workspace, "get_groups", check_permissions)
    except ResourceExceptions.UserNotAuthorized:
        await Permissions.check_permissions(group, "get_group", check_permissions)

    members = (await actions.MembersActions.get_members(group)).members if include_members else None
    policies = (await actions.PolicyActions.get_policies(resource=group)).policies if include_policies else None
    workspace = WorkspaceSchemas.Workspace(**group.workspace.model_dump(exclude={"members",  # type: ignore
                                                                                 "policies",
                                                                                 "groups"}))
    # Return the workspace with the fetched resources
    return GroupSchemas.Group(id=group.id,
                              name=group.name,
                              description=group.description,
                              workspace=workspace,
                              members=members,
                              policies=policies)


# Update a group
async def update_group(group: Group,
                       group_data: GroupSchemas.GroupUpdateRequest,
                       check_permissions: bool = True) -> GroupSchemas.Group:
    try:
        await Permissions.check_permissions(group.workspace, "update_groups", check_permissions)
    except WorkspaceExceptions.UserNotAuthorized:
        await Permissions.check_permissions(group, "update_group", check_permissions)

    save_changes = False
    workspace: Workspace = group.workspace  # type: ignore
    # The group must belong to a workspace
    if not workspace:
        raise WorkspaceExceptions.WorkspaceNotFound(workspace)

    # Check if group name is provided
    if group_data.name and group_data.name != group.name:
        # Check if group name is unique
        for g in workspace.groups:
            if g.name == group_data.name:  # type: ignore
                raise GroupExceptions.NonUniqueName(group)
        group.name = group_data.name  # Update the group name
        save_changes = True
    # Check if group description is provided
    if group_data.description and group_data.description != group.description:
        group.description = group_data.description  # Update the group description
        save_changes = True

    # Save the updates
    if save_changes:
        await Group.save(group)
    # Return the updated group
    return GroupSchemas.Group(**group.model_dump())


# Delete a group
async def delete_group(group: Group,
                       check_permissions: bool = True):
    try:
        await Permissions.check_permissions(group.workspace, "delete_groups", check_permissions)
    except WorkspaceExceptions.UserNotAuthorized:
        await Permissions.check_permissions(group, "delete_group", check_permissions)

    workspace: Workspace = group.workspace  # type: ignore
    group_ref = DBRef(collection="Group", id=group.id)

    # Remove the group from group list in the workspace
    workspace.groups = [g for g in workspace.groups if g.id != group.id]  # type: ignore
    await Workspace.replace(workspace)

    # Delete the group
    await Group.delete(group)
    if await Group.get(group.id):
        raise GroupExceptions.ErrorWhileDeleting(group.id)

    # Delete group policy in the workspace and all policies stored in the group
    # MongoDB style: {"$or": [{"parent_resource": ref}, {"policy_holder": ref}]}
    # Beanie style: Or(Policy.parent_resource == ref, Policy.policy_holder == ref)
    await Policy.find(Or(Policy.parent_resource == group_ref,          # type: ignore
                         Policy.policy_holder == group_ref)).delete()  # type: ignore



from colorama import Fore
from typing import List


# Print colored messages
def print_in_color(
    *message: List[str] | str,
    color: str,
    type: str,
        source: str = "") -> None:
    str_message = ' '.join([str(i) for i in message])
    print("{color}{type}{source}{message}".format(
        type=type + Fore.RESET + ":" + (9 - len(type)) * " ",
        color=color,
        source=source,
        message=str_message))


# For general information
def info(*message: List[str] | str) -> None:
    print_in_color(*message, color=Fore.GREEN, type="INFO")


# For final result of a test
def test_success(*message: List[str] | str) -> None:
    print_in_color(*message, color=Fore.GREEN, type="SUCCESS")


# For general test information
def test_info(*message: List[str] | str) -> None:
    print_in_color(*message, color=Fore.BLUE, type="TESTING")


# Print message with red ERROR header
# Used for critical errors
def print_error(*message: List[str] | str) -> None:
    print_in_color(*message, color=Fore.RED, type="ERROR")


# Used for non-critical errors that won't affect app performance
def print_warning(*message: List[str] | str) -> None:
    print_in_color(*message, color=Fore.YELLOW, type="WARNING")


# Print message with blue MESSAGE header
def print_message(*message: List[str] | str, source: str = "") -> None:
    print_in_color(
        *message,
        color=Fore.BLUE,
        type="MESSAGE",
        source=f"from {source}: ")



# FastAPI
from typing import Annotated, Literal
from fastapi import APIRouter, Body, Depends, HTTPException, Query, status
from unipoll_api import dependencies as Dependencies
from unipoll_api.actions import GroupActions, PermissionsActions, MembersActions, PolicyActions
from unipoll_api.exceptions.resource import APIException
from unipoll_api.schemas import GroupSchemas, PolicySchemas, MemberSchemas
from unipoll_api.documents import Account, Group, Policy, ResourceID


# APIRouter creates path operations for user module
open_router: APIRouter = APIRouter()
# router: APIRouter = APIRouter(dependencies=[Depends(Dependencies.check_group_permission)])
router: APIRouter = APIRouter()


# Get groups
@open_router.get("/", response_description="List of groups")
async def get_all_groups(workspace: Annotated[ResourceID | None, Query()] = None,
                         account: Annotated[ResourceID | None, Query()] = None,
                         name: Annotated[str | None, Query()] = None
                         ) -> GroupSchemas.GroupList:
    return await GroupActions.get_groups(workspace=await Dependencies.get_workspace(workspace) if workspace else None,
                                         account=await Dependencies.get_account(account) if account else None,
                                         name=name)


# Create a new group
@open_router.post("/",
                  status_code=201,
                  response_description="Created Group",
                  response_model=GroupSchemas.GroupCreateOutput)
async def create_group(input_data: GroupSchemas.GroupCreateRequest = Body(...)):
    try:
        workspace = await Dependencies.get_workspace(input_data.workspace)
        return await GroupActions.create_group(workspace, name=input_data.name, description=input_data.description)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


query_params = list[Literal["policies", "members", "all"]]


# Get group info by id
@router.get("/{group_id}",
            response_description="Get a group",
            response_model=GroupSchemas.Group,
            response_model_exclude_defaults=True,
            response_model_exclude_none=True)
async def get_group(group: Group = Depends(Dependencies.get_group),
                    include: Annotated[query_params | None, Query()] = None):
    try:
        params = {}
        if include:
            if "all" in include:
                params = {"include_members": True, "include_policies": True}
            else:
                if "members" in include:
                    params["include_members"] = True
                if "policies" in include:
                    params["include_policies"] = True
        return await GroupActions.get_group(group, **params)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Update group info
@router.patch("/{group_id}",
              response_description="Update a group",
              response_model=GroupSchemas.GroupShort)
async def update_group(group_data: GroupSchemas.GroupUpdateRequest,
                       group: Group = Depends(Dependencies.get_group)):
    try:
        return await GroupActions.update_group(group, group_data)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Delete a group
@router.delete("/{group_id}",
               status_code=status.HTTP_204_NO_CONTENT,
               response_description="Delete a group")
async def delete_group(group: Group = Depends(Dependencies.get_group)):
    try:
        await GroupActions.delete_group(group)
        return status.HTTP_204_NO_CONTENT
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Get a list of group members
@router.get("/{group_id}/members",
            response_description="List of group members",
            response_model=MemberSchemas.MemberList,
            response_model_exclude_unset=True)
async def get_group_members(group: Group = Depends(Dependencies.get_group)):
    try:
        return await MembersActions.get_members(group)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Add member to group
@router.post("/{group_id}/members",
             response_description="List of group members",
             response_model=MemberSchemas.MemberList)
async def add_group_members(member_data: MemberSchemas.AddMembers,
                            group: Group = Depends(Dependencies.get_group)):
    try:
        return await MembersActions.add_members(group, member_data.accounts)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Remove members from the workspace
@router.delete("/{group_id}/members/{account_id}",
               response_description="Updated list removed members",
               response_model_exclude_unset=True)
async def remove_group_member(group: Group = Depends(Dependencies.get_group),
                              account: Account = Depends(Dependencies.get_account)):
    try:
        return await MembersActions.remove_member(group, account)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# List all policies in the workspace
@router.get("/{group_id}/policies",
            response_description="List of all policies",
            response_model=PolicySchemas.PolicyList)
async def get_group_policies(group: Group = Depends(Dependencies.get_group),
                             account_id: ResourceID = Query(None)) -> PolicySchemas.PolicyList:
    try:
        account = await Dependencies.get_account(account_id) if account_id else None
        member = await Dependencies.get_member(account, group) if account else None
        return await PolicyActions.get_policies(resource=group, policy_holder=member)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Set permissions for a user in a group
@router.put("/{group_id}/policies/{policy_id}",
            response_description="Updated policy",
            response_model=PolicySchemas.PolicyOutput)
async def set_group_policy(group: Group = Depends(Dependencies.get_group),
                           policy: Policy = Depends(Dependencies.get_policy),
                           permissions: PolicySchemas.PolicyInput = Body(...)):
    """
    Sets the permissions for a user in a workspace.
    Query parameters:
        @param workspace_id: id of the workspace to update
    Body parameters:
    - **user_id** (str): id of the user to update
    - **permissions** (int): new permissions for the user
    """
    try:
        return await PolicyActions.update_policy(policy, new_permissions=permissions.permissions)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Get All Group Permissions
@open_router.get("/permissions",
                 response_description="List of all Group permissions",
                 response_model=PolicySchemas.PermissionList)
async def get_group_permissions():
    try:
        return await PermissionsActions.get_group_permissions()
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))
